package me.chyxion.tigon.mybatis.test;

import lombok.extern.slf4j.Slf4j;
import me.chyxion.tigon.mybatis.test.model.Activity;
import org.apache.commons.lang3.StringUtils;
import org.junit.Test;
import org.springframework.data.domain.Sort;
import org.springframework.data.mapping.PropertyPath;
import org.springframework.data.repository.query.parser.Part;
import static org.springframework.data.repository.query.parser.Part.Type.*;
import org.springframework.data.repository.query.parser.PartTree;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;
import javax.xml.parsers.*;
import java.io.IOException;
import java.util.*;

import org.xml.sax.helpers.DefaultHandler;
import me.chyxion.tigon.mybatis.BaseMapper;
import org.apache.ibatis.parsing.XPathParser;
import java.io.UnsupportedEncodingException;
import javax.xml.transform.TransformerException;
import me.chyxion.tigon.mybatis.test.model.App;
import org.springframework.core.GenericTypeResolver;
import me.chyxion.tigon.mybatis.test.mapper.AppMapper;
import org.apache.ibatis.builder.xml.XMLMapperEntityResolver;

/**
 * @version 0.0.1
 * @since 0.0.1
 * @author Shaun Chyxion <br>
 * chyxion@163.com <br>
 * May 13, 2016 10:45:07 AM
 */
@Slf4j
public class TestDriver {

    @Test
    public void testMapperParse() throws TransformerException, UnsupportedEncodingException {
        Map<String, Object> stringObjectMap = new App().toMap();
        System.err.println();
        System.err.println(new App().cols());
    }

    @Test
    public void testGetModel() {
        Class<?>[] classes = GenericTypeResolver.resolveTypeArguments(AppMapper.class, BaseMapper.class);
        System.err.println(classes);
    }

    @Test
    public void testParseXml() throws ParserConfigurationException, IOException, SAXException {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setNamespaceAware(true);
        DocumentBuilder builder = factory.newDocumentBuilder();
        long timeStart = System.currentTimeMillis();
        Document doc = builder.parse(TestDriver.class
            .getResourceAsStream("/mybatis/mappers/app-mapper.xml"));
        long timeElapse = System.currentTimeMillis() - timeStart;
        System.err.println("Time Elapse: " + timeElapse);
    }

    @Test
    public void testParseXml2() {
        long timeStart = System.currentTimeMillis();
        XPathParser xPathParser = new XPathParser(TestDriver.class
                .getResourceAsStream("/mybatis/mappers/app-mapper.xml"),
                false, null, new XMLMapperEntityResolver());
        long timeElapse = System.currentTimeMillis() - timeStart;
        System.err.println("Time Elapse: " + timeElapse);
    }

    @Test
    public void testParseXml3() throws ParserConfigurationException, SAXException, IOException {
        long timeStart = System.currentTimeMillis();
        //创建一个解析XML的工厂对象
        SAXParserFactory parserFactory=SAXParserFactory.newInstance();
        //创建一个解析XML的对象
        SAXParser parser = parserFactory.newSAXParser();
        parser.parse(TestDriver.class
            .getResourceAsStream("/mybatis/mappers/app-mapper.xml"),
                new DefaultHandler());
        long timeElapse = System.currentTimeMillis() - timeStart;
        System.err.println("Time Elapse: " + timeElapse);
    }

    @Test
    public void testParse() {
        final PartTree tree = new PartTree("findByIdLessThanAndIdStartsWithAndIdInAndIdNotInOrIdAndHoldingDateOrIdBetweenOrderById", Activity.class);
        final Iterator<PartTree.OrPart> itOrParts = tree.iterator();

        List<StringBuilder> sqlFrags = new LinkedList<StringBuilder>();
        int paramIndex = 0;
        while (itOrParts.hasNext()) {
            final PartTree.OrPart orPart = itOrParts.next();
            log.info("Or part [{}] found.", orPart);
            final Iterator<Part> itOrPart = orPart.iterator();
            List<String> andParts = new LinkedList<String>();
            int orPartIndex = 0;
            while (itOrPart.hasNext()) {
                final Part part = itOrPart.next();
                final PropertyPath prop = part.getProperty();
                final Part.Type type = part.getType();
                final int numArgs = part.getNumberOfArguments();
                andParts.add(buildSQL(part, paramIndex, orPartIndex == 0));
                log.info("Part [{}], path [{}], number of arg [{}], type [{}].", part, prop.toDotPath(), numArgs, type.name());
                paramIndex += numArgs;
                ++orPartIndex;
            }
            log.info("[{}] and parts found.", andParts.size());
            sqlFrags.add(new StringBuilder(StringUtils.join(andParts, " ")));
        }
        log.info("SQL [{}].", StringUtils.join(sqlFrags, " "));
        final Sort sort = tree.getSort();
        if (sort != null) {
            Iterator<Sort.Order> itSort = sort.iterator();
            while (itSort.hasNext()) {
                Sort.Order order = itSort.next();
                log.info("Order [{}] [{}]", order.getProperty(), order.getDirection());
            }
        }
    }

    private String buildSQL(Part part, final int paramIndex, boolean orClause) {
        final PropertyPath prop = part.getProperty();
        final Part.Type type = part.getType();
        final String col = prop.toDotPath();

        if (BETWEEN.equals(type)) {
            //
            return col + " between #{param" + paramIndex + "} and #{param" + (paramIndex + 1) + "}";
        }
        else if (CONTAINING.equals(type)) {
            return ifNotNullTag(paramIndex, false, orClause, true, col + " like #{param" + paramIndex + "}");
        }
        else if (ENDING_WITH.equals(type)) {
            return ifNotNullTag(paramIndex, false, orClause, true, col + " like #{param" + paramIndex + "}");
        }
        else if (FALSE.equals(type)) {
            return ifNotNullTag(paramIndex, false, orClause, false, col + " is false");
        }
        else if (GREATER_THAN.equals(type) || AFTER.equals(type)) {
            return ifNotNullTag(paramIndex, false, orClause, true, col + " > #{param" + paramIndex + "}");
        }
        else if (GREATER_THAN_EQUAL.equals(type)) {
            return ifNotNullTag(paramIndex, false, orClause, true, col + " >= #{param" + paramIndex + "}");
        }
        else if (IN.equals(type)) {
            return ifNotNullTag(paramIndex, true, orClause, true, col + " in " + buildArray(paramIndex));
        }
        else if (IS_NOT_NULL.equals(type) || EXISTS.equals(type)) {
            return ifNotNullTag(paramIndex, false, orClause, false, col + " is not null");
        }
        else if (IS_NULL.equals(type)) {
            return ifNotNullTag(paramIndex, false, orClause, false, col + " is null");
        }
        else if (LESS_THAN.equals(type) || BEFORE.equals(type)) {
            return ifNotNullTag(paramIndex, false, orClause, true, col + " < #{param" + paramIndex + "}");
        }
        else if (LESS_THAN_EQUAL.equals(type)) {
            return ifNotNullTag(paramIndex, false, orClause, true, col + " <= #{param" + paramIndex + "}");
        }
        else if (LIKE.equals(type)) {
            return ifNotNullTag(paramIndex, false, orClause, true, col + " like #{param" + paramIndex + "}");
        }
        else if (NEGATING_SIMPLE_PROPERTY.equals(type)) {
            return col + " <> #{param" + paramIndex + "}";
        }
        else if (NOT_IN.equals(type)) {
            return ifNotNullTag(paramIndex, true, orClause, true, col + " not in " + buildArray(paramIndex));
        }
        else if (NOT_LIKE.equals(type)) {
            return ifNotNullTag(paramIndex, false, orClause, true, col + " not like #{param" + paramIndex + "}");
        }
        else if (REGEX.equals(type)) {
            return ifNotNullTag(paramIndex, false, orClause, true, col + " regexp #{param" + paramIndex + "}");
        }
        else if (SIMPLE_PROPERTY.equals(type)) {
            return ifNotNullTag(paramIndex, false, orClause, true, col + " = #{param" + paramIndex + "}");
        }
        else if (STARTING_WITH.equals(type)) {
            return ifNotNullTag(paramIndex, false, orClause, true,
                bindTag("param" + paramIndex + "_var", "param" + paramIndex + " + '%'") +
                    col + " like #{param" + paramIndex + "_var}");
        }
        else if (TRUE.equals(type)) {
            return ifNotNullTag(paramIndex, false, orClause, false, col + " is true");
        }
        else if (NOT_CONTAINING.equals(type)) {
            // TODO
            return col + " not like #{param" + paramIndex + "}";
        }
        // redis
        else if (NEAR.equals(type)) {
        }
        // redis
        else if (WITHIN.equals(type)) {
            return col + " like #{param" + paramIndex + "}";
        }
        return null;
    }

    String ifNotNullTag(int paramIndex, boolean collection, boolean orClause, boolean wrapIf, String sql) {
        if (wrapIf) {
            return ifNotNullTag(paramIndex, collection, orClause, sql);
        }
        return (orClause ? " or " : " and ") + sql;
    }

    String bindTag(final String name, final String value) {
        return "<bind name=\"" + name + "\" value=\"" + value + "\" />";
    }

    String buildArray(int paramIndex) {
        return new StringBuilder("<foreach collection=\"")
            .append("param")
            .append(paramIndex)
            .append("\" item=\"__val__\" open=\"(\" separator=\", \" close=\")\">\n")
            .append("#{__val__}")
            .append("</foreach>\n")
            .toString();
    }

    String ifNotNullTag(int paramIndex, boolean orClause, String innerTag) {
        return new StringBuilder("<if test=\"").append("param")
            .append(paramIndex)
            .append(" != null\">\n")
            .append(orClause ? " or " : " and ")
            .append(innerTag)
            .append("</if>\n")
            .toString();
    }

    String ifNotEmptyTag(int paramIndex, boolean orClause, String innerTag) {
        return new StringBuilder("<if test=\"")
            .append("param")
            .append(paramIndex)
            .append(" != null and ((")
            .append("param")
            .append(paramIndex)
            .append(".getClass().isArray() and ")
            .append("param")
            .append(paramIndex)
            .append(".length")
            .append(") or (")
            .append("param")
            .append(paramIndex)
            .append(" instanceof java.lang.Iterable and ")
            .append("param")
            .append(paramIndex)
            .append(".iterator().hasNext()")
            .append("))\">\n")
            .append(orClause ? " or " : " and ")
            .append(innerTag)
            .append("</if>\n")
            .toString();
    }

    String ifNotNullTag(int paramIndex, boolean collection, boolean orClause, String innerTag) {
        StringBuilder sbRtn = new StringBuilder("<if test=\"");
        if (collection) {
            sbRtn.append("param")
                    .append(paramIndex)
                    .append(" != null and ((")
                    .append("param")
                    .append(paramIndex)
                    .append(".getClass().isArray() and ")
                    .append("param")
                    .append(paramIndex)
                    .append(".length")
                    .append(") or (")
                    .append("param")
                    .append(paramIndex)
                    .append(" instanceof java.lang.Iterable and ")
                    .append("param")
                    .append(paramIndex)
                    .append(".iterator().hasNext()")
                    .append("))\">\n");
        }
        else {
            sbRtn.append("param")
                    .append(paramIndex)
                    .append(" != null\">\n");
        }
        return sbRtn.append(orClause ? " or " : " and ")
                .append(innerTag)
                .append("</if>\n").toString();
    }

    String ifNotNullTag(int paramIndex, boolean collection, boolean orClause, InnerTag innerTag) {
        StringBuilder sbRtn = new StringBuilder("<if test=\"");
        if (collection) {
            sbRtn.append("param")
                .append(paramIndex)
                .append(" != null and ((")
                .append("param")
                .append(paramIndex)
                .append(".getClass().isArray() and ")
                .append("param")
                .append(paramIndex)
                .append(".length")
                .append(") or (")
                .append("param")
                .append(paramIndex)
                .append(" instanceof java.lang.Iterable and ")
                .append("param")
                .append(paramIndex)
                .append(".iterator().hasNext()")
                .append("))\">\n");
        }
        else {
            sbRtn.append("param")
                .append(paramIndex)
                .append(" != null\">\n");
        }
        return sbRtn.append(orClause ? " or " : " and ")
            .append(innerTag.tag())
            .append("</if>\n").toString();
    }

    interface InnerTag {
        String tag();
    }

    @Test
    public void testParse1() {
        PartTree tree = new PartTree("findIdById", Activity.class);
        log.info("tree to string: [{}].", tree);
        Iterator<PartTree.OrPart> iterator = tree.iterator();
        while (iterator.hasNext()) {
            PartTree.OrPart next = iterator.next();
            log.info("OrPart [{}].", next);
            Iterator<Part> iterator1 = next.iterator();
            while (iterator1.hasNext()) {
                Part next1 = iterator1.next();

                PropertyPath property = next1.getProperty();
                Part.Type type = next1.getType();
                Collection<String> keywords = type.getKeywords();
                log.info("Part [{}], path [{}], number of arg [{}], type [{}].", next1, property.toDotPath(), next1.getNumberOfArguments(), type.name());
            }
        }

        log.info("join or part [{}].", StringUtils.join(tree.iterator(), " or "));
        System.err.println("");

        for (Part part : tree.getParts()) {
            System.err.println(part);
        }
        // System.err.println(tree);

//        Iterator<Sort.Order> iterator1 = tree.getSort().iterator();
//        while (iterator1.hasNext()) {
//            System.err.println(iterator1.next());
//        }
        // tree = new PartTree("listByIdOrLastName", Customer.class);
        // System.err.println(tree);
    }
}
